from libcamera import ColorSpace, Rectangle, Size

import picamera2.formats as formats


def convert_from_libcamera_type(value):
    if isinstance(value, Rectangle):
        value = (value.x, value.y, value.width, value.height)
    elif isinstance(value, Size):
        value = (value.width, value.height)
    elif isinstance(value, (list, tuple)) and all(isinstance(item, Rectangle) for item in value):
        value = [(v.x, v.y, v.width, v.height) for v in value]
    return value


def colour_space_to_libcamera(colour_space, format):
    # libcamera may complain if we supply an RGB format stream with a YCbCr matrix or range.
    if formats.is_RGB(format):
        colour_space = ColorSpace(colour_space)  # it could be shared with other streams, so copy it
        colour_space.ycbcrEncoding = ColorSpace.YcbcrEncoding.Null
        colour_space.range = ColorSpace.Range.Full
    return colour_space


COLOUR_SPACE_TABLE = {ColorSpace.Sycc(), ColorSpace.Smpte170m(), ColorSpace.Rec709()}


def colour_space_from_libcamera(colour_space):
    # Colour spaces may come back from libcamera without a YCbCr matrix or range, meaning
    # they don't look like the 3 standard colour spaces (in the table) that we expect people
    # to use. Let's fix that.
    if colour_space is None:  # USB webcams might have a "None" colour space
        return None
    for cs in COLOUR_SPACE_TABLE:
        if colour_space.primaries == cs.primaries and colour_space.transferFunction == cs.transferFunction:
            return cs
    return colour_space



"""This is a base class for a multi-threaded software encoder."""

from fractions import Fraction
from math import sqrt

import av

from picamera2.encoders.encoder import Encoder, Quality

from ..request import MappedArray


class LibavH264Encoder(Encoder):
    """Encoder class that uses libx264 for h.264 encoding."""

    def __init__(self, bitrate=None, repeat=True, iperiod=30, framerate=30, qp=None):
        """Initialise"""
        super().__init__()
        self._codec = "h264"  # for now only support h264
        self.repeat = repeat
        self.bitrate = bitrate
        self.iperiod = iperiod
        self.framerate = framerate
        self.qp = qp

    def _setup(self, quality):
        # If an explicit quality was specified, use it, otherwise try to preserve any bitrate/qp
        # the user may have set for themselves.
        if quality is not None or \
           (getattr(self, "bitrate", None) is None and getattr(self, "qp", None) is None):
            quality = Quality.MEDIUM if quality is None else quality
            # These are suggested bitrates for 1080p30 in Mbps
            BITRATE_TABLE = {Quality.VERY_LOW: 3,
                             Quality.LOW: 4,
                             Quality.MEDIUM: 7,
                             Quality.HIGH: 10,
                             Quality.VERY_HIGH: 14}
            reference_complexity = 1920 * 1080 * 30
            actual_complexity = self.width * self.height * getattr(self, "framerate", 30)
            reference_bitrate = BITRATE_TABLE[quality] * 1000000
            self.bitrate = int(reference_bitrate * sqrt(actual_complexity / reference_complexity))

    def _start(self):
        self._container = av.open("/dev/null", "w", format="null")
        self._stream = self._container.add_stream(self._codec, rate=self.framerate)

        self._stream.codec_context.thread_count = 8
        self._stream.codec_context.thread_type = av.codec.context.ThreadType.FRAME

        self._stream.width = self.width
        self._stream.height = self.height
        self._stream.pix_fmt = "yuv420p"

        if self.bitrate is not None:
            self._stream.codec_context.bit_rate = self.bitrate
        self._stream.codec_context.gop_size = self.iperiod
        self._stream.codec_context.options["preset"] = "ultrafast"
        self._stream.codec_context.options["deblock"] = "1"
        # Absence of the "global header" flags means that SPS/PPS headers get repeated.
        if not self.repeat:
            self._stream.codec_context.flags |= av.codec.context.Flags.GLOBAL_HEADER
        if self.qp is not None:
            self._stream.codec_context.qmin = self.qp
            self._stream.codec_context.qmax = self.qp

        self._stream.codec_context.time_base = Fraction(1, 1000000)

        FORMAT_TABLE = {"YUV420": "yuv420p",
                        "BGR888": "rgb24",
                        "RGB888": "bgr24",
                        "XBGR8888": "rgba",
                        "XRGB8888": "bgra"}
        self._av_input_format = FORMAT_TABLE[self._format]

    def _stop(self):
        for packet in self._stream.encode():
            self.outputframe(bytes(packet), packet.is_keyframe, timestamp=packet.pts)
        self._container.close()

    def _encode(self, stream, request):
        timestamp_us = self._timestamp(request)
        with MappedArray(request, stream) as m:
            frame = av.VideoFrame.from_ndarray(m.array, format=self._av_input_format, width=self.width)
            frame.pts = timestamp_us
            for packet in self._stream.encode(frame):
                self.outputframe(bytes(packet), packet.is_keyframe, timestamp=packet.pts)



# Installing Qt and OpenGL on the 64-bit Lite OS, and trying to run
# a remote preview window, causes an error here (the 32-bit Lite OS is OK).
# It may be something to do with more recent versions of python3-opengl?
# Anyway, if we carry on regardless at least the non-OpenGL preview works,
# which is in any case what is required for remote preview windows.
from logging import getLogger

from .q_picamera2 import QPicamera2

_log = getLogger(__name__)

try:
    from .q_gl_picamera2 import QGlPicamera2
except Exception:
    _log.warning("OpenGL will not be available")



#!/usr/bin/python3
import threading

from libcamera import ControlType, Rectangle, Size


class Controls():
    def _framerates_to_durations_(framerates):
        if not isinstance(framerates, (tuple, list)):
            framerates = (framerates, framerates)
        return (int(1000000 / framerates[1]), int(1000000 / framerates[0]))

    def _durations_to_framerates_(durations):
        if durations[0] == durations[1]:
            return 1000000 / durations[0]
        return (1000000 / durations[1], 1000000 / durations[0])

    _VIRTUAL_FIELDS_MAP_ = {"FrameRate": ("FrameDurationLimits", _framerates_to_durations_, _durations_to_framerates_)}

    def __init__(self, picam2, controls={}):
        self._picam2 = picam2
        self._controls = []
        self._lock = threading.Lock()
        self.set_controls(controls)

    def __setattr__(self, name, value):
        if not name.startswith('_'):
            if name in Controls._VIRTUAL_FIELDS_MAP_:
                real_field = Controls._VIRTUAL_FIELDS_MAP_[name]
                name = real_field[0]
                value = real_field[1](value)
            if name not in self._picam2.camera_ctrl_info.keys():
                raise RuntimeError(f"Control {name} is not advertised by libcamera")
            self._controls.append(name)
        self.__dict__[name] = value

    def __getattribute__(self, name):
        if name in Controls._VIRTUAL_FIELDS_MAP_:
            real_field = Controls._VIRTUAL_FIELDS_MAP_[name]
            real_value = self.__getattribute__(real_field[0])
            return real_field[2](real_value)
        return super().__getattribute__(name)

    def __repr__(self):
        return f"<Controls: {self.make_dict()}>"

    def __enter__(self):
        self._lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self._lock.release()

    def set_controls(self, controls):
        with self._lock:
            if isinstance(controls, dict):
                for k, v in controls.items():
                    self.__setattr__(k, v)
            elif isinstance(controls, Controls):
                for k in controls._controls:
                    v = controls.__dict__[k]
                    self.__setattr__(k, v)
            else:
                raise RuntimeError(f"Cannot update controls with {type(controls)} type")

    def get_libcamera_controls(self):
        def list_or_tuple(thing):
            return type(thing) in {list, tuple}

        libcamera_controls = {}
        with self._lock:
            for k in self._controls:
                v = self.__dict__[k]
                id = self._picam2.camera_ctrl_info[k][0]
                if id.type == ControlType.Rectangle:
                    # We can get a list of Rectangles or a single one.
                    if list_or_tuple(v) and v and list_or_tuple(v[0]):
                        v = [Rectangle(*i) for i in v]
                    else:
                        v = Rectangle(*v)
                elif id.type == ControlType.Size:
                    v = Size(*v)
                libcamera_controls[id] = v
        return libcamera_controls

    def make_dict(self):
        dict_ = {}
        with self._lock:
            for k in self._controls:
                v = self.__dict__[k]
                dict_[k] = v
        return dict_
